/*****************************************************************************
 * $CAMITK_LICENCE_BEGIN$
 *
 * CamiTK - Computer Assisted Medical Intervention ToolKit
 * (c) 2001-2024 Univ. Grenoble Alpes, CNRS, Grenoble INP - UGA, TIMC, 38000 Grenoble, France
 *
 * Visit http://camitk.imag.fr for more information
 *
 * This file is part of CamiTK.
 *
 * CamiTK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * CamiTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with CamiTK.  If not, see <http://www.gnu.org/licenses/>.
 *
 * $CAMITK_LICENCE_END$
 ****************************************************************************/

// - Qt stuff
#include <QColor>
#include <QColorDialog>

// - Local stuff
#include "ColorPointWidget.h"
#include "VolumeRenderingWidget.h"

TransparencyPointWidget::TransparencyPointWidget(QWidget* volumeRenderingWidget) : QWidget(volumeRenderingWidget) {
    ui.setupUi(this);
    this->volumeRenderingWidget = volumeRenderingWidget;
}

TransparencyPointWidget::TransparencyPointWidget(QWidget* volumeRenderingWidget, double grayLevel, double transparency) : QWidget(volumeRenderingWidget) {
    ui.setupUi(this);
    this->volumeRenderingWidget = volumeRenderingWidget;

    ui.doubleSpinBox->setValue(grayLevel);
    ui.transparencySpinBox->setValue(transparency);

}

void TransparencyPointWidget::remove() {
    auto* dadVR = dynamic_cast<VolumeRenderingWidget*>(this->volumeRenderingWidget);
    if (dadVR) {
        dadVR->removeTransparencyPoint(this);
    }
}

double TransparencyPointWidget::getGrayLevel() {
    return ui.doubleSpinBox->value();
}

double TransparencyPointWidget::getTransparency() {
    return ui.transparencySpinBox->value();
}


