/*****************************************************************************
 * $CAMITK_LICENCE_BEGIN$
 *
 * CamiTK - Computer Assisted Medical Intervention ToolKit
 * (c) 2001-2024 Univ. Grenoble Alpes, CNRS, Grenoble INP - UGA, TIMC, 38000 Grenoble, France
 *
 * Visit http://camitk.imag.fr for more information
 *
 * This file is part of CamiTK.
 *
 * CamiTK is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * CamiTK is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with CamiTK.  If not, see <http://www.gnu.org/licenses/>.
 *
 * $CAMITK_LICENCE_END$
 ****************************************************************************/

#ifndef CONTOURWIDGETVTKCOMMAND_H
#define CONTOURWIDGETVTKCOMMAND_H

#include <vtkCommand.h>

#include <QCoreApplication>

class ContourWidget;


/**
 * Vtk Command that control what is done when the widget is modified interactively by the user
 */
class ContourWidgetVtkCommand : public vtkCommand {
    Q_DECLARE_TR_FUNCTIONS(ContourWidgetVtkCommand)
public:
    /// the constructor
    ContourWidgetVtkCommand(ContourWidget* a);

    /** Allows to get the current 3D widget configuration.
     *  This is a VTK "callback" (inherited from vtkCommand).
     */
    void Execute(vtkObject* caller, unsigned long, void*) override;

private:
    ContourWidget* contourWidgetAction;
};

#endif // CONTOURWIDGETVTKCOMMAND_H
